@@ -3,7 +3,7 @@
  *
  *      Copyright 2008 - 2010 PCMan <pcman.tw@gmail.com>
  *      Copyright 2009 Jürgen Hötzel <juergen@archlinux.org>
- *      Copyright 2012-2015 Andriy Grytsenko (LStranger) <andrej@rep.kiev.ua>
+ *      Copyright 2012-2017 Andriy Grytsenko (LStranger) <andrej@rep.kiev.ua>
  *      Copyright 2016 Mamoru TASAKA <mtasaka@fedoraproject.org>
  *
  *      This file is a part of libmenu-cache package and created program
@@ -34,6 +34,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <glib.h>
+#include <glib/gstdio.h>
 #include <gio/gio.h>
 #include <sys/types.h>
 #include <sys/socket.h>
@@ -45,6 +46,7 @@
 #include <errno.h>
 #include <signal.h>
 #include <utime.h>
+#include <sys/wait.h>
 
 #ifdef G_ENABLE_DEBUG
 #define DEBUG(...)  g_debug(__VA_ARGS__)
@@ -131,12 +133,12 @@ static gboolean delayed_cache_free(gpointer data)
 
 static void cache_free(Cache* cache)
 {
-    /* shutdown cache in 10 minutes of inactivity */
+    /* shutdown cache in 6 seconds of inactivity */
     if(!cache->delayed_free_handler)
-        cache->delayed_free_handler = g_timeout_add_seconds(600,
+        cache->delayed_free_handler = g_timeout_add_seconds(6,
                                                             delayed_cache_free,
                                                             cache);
-    DEBUG("menu %p cache unused, removing in 600s", cache);
+    DEBUG("menu %p cache unused, removing in 6s", cache);
 }
 
 static gboolean read_all_used_files( FILE* f, int* n_files, char*** used_files )
@@ -221,6 +223,7 @@ static gboolean regenerate_cache( const char* menu_name,
     FILE* f;
     int n_files, status = 0;
     char** files;
+    const char *user_data_dir = env[5];
     const char* argv[] = {
         MENUCACHE_LIBEXECDIR "/menu-cache-gen",
         "-l", NULL,
@@ -233,6 +236,18 @@ static gboolean regenerate_cache( const char* menu_name,
 
     /* DEBUG("cmd: %s", g_strjoinv(" ", argv)); */
 
+    /* create $XDG_DATA_HOME/applications if it does not exist yet */
+    if (!user_data_dir || !user_data_dir[0])
+        user_data_dir = g_get_user_data_dir();
+    if (g_file_test(user_data_dir, G_FILE_TEST_IS_DIR) ||
+        g_mkdir(user_data_dir, 0700) == 0)
+    {
+        char *local_app_path = g_build_filename(user_data_dir, "applications", NULL);
+        if (!g_file_test(local_app_path, G_FILE_TEST_IS_DIR))
+            g_mkdir(local_app_path, 0700);
+        g_free(local_app_path);
+    }
+
     /* run menu-cache-gen */
     /* FIXME: is cast to (char**) valid here? */
     if( !g_spawn_sync(NULL, (char **)argv, NULL, 0,
@@ -376,25 +391,39 @@ void on_file_changed( GFileMonitor* mon, GFile* gf, GFile* other,
         if( G_LIKELY(idx < cache->n_files) && cache->files[idx][0] == 'D' )
         {
             char* changed_file = g_file_get_path(gf);
-            char* dir_path = cache->files[idx]+1;
-            int len = strlen(dir_path);
-            /* if the changed file is a file in the monitored dir */
-            if( strncmp(changed_file, dir_path, len) == 0 && changed_file[len] == '/' )
+            /* Regenerate the cache if the changed file is a directory.
+             *
+             * The file monitor isn't recursive, so imagine we add a
+             * subdirectory to /usr/share/applications, and subsequently
+             * add a desktop entry to that. If we ignore the new
+             * subdirectory, we won't notice when the desktop entry is
+             * added. By regenerating the cache, the subdirectory will
+             * be mentioned there, picked up by read_all_used_files(),
+             * and monitored for subsequent changes.
+             */
+            if (!g_file_test(changed_file, G_FILE_TEST_IS_DIR))
             {
-                char* base_name = changed_file + len + 1;
-                gboolean skip = TRUE;
-                /* only *.desktop and *.directory files can affect the content of the menu. */
-                if( g_str_has_suffix(base_name, ".desktop") )
+                char* dir_path = cache->files[idx]+1;
+                int len = strlen(dir_path);
+                /* if the changed file is a file in the monitored dir */
+                if( strncmp(changed_file, dir_path, len) == 0 && changed_file[len] == '/' )
                 {
+                    char* base_name = changed_file + len + 1;
+                    gboolean skip = TRUE;
+                    /* only *.desktop and *.directory files can affect the content of the menu. */
+                    if( g_str_has_suffix(base_name, ".desktop") )
+                    {
+                            skip = FALSE;
+                    }
+                    else if( g_str_has_suffix(base_name, ".directory") )
                         skip = FALSE;
-                }
-                else if( g_str_has_suffix(base_name, ".directory") )
-                    skip = FALSE;
 
-                if( skip )
-                {
-                    DEBUG("files are changed, but no re-generation is needed.");
-                    return;
+                    if( skip )
+                    {
+                        DEBUG("files are changed, but no re-generation is needed.");
+                        g_free(changed_file);
+                        return;
+                    }
                 }
             }
             g_free(changed_file);
@@ -473,21 +502,33 @@ static void get_socket_name( char* buf, int len )
         if(*p)
             *p = '\0';
     }
+    /* NOTE: this socket name is incompatible with versions > 1.0.2,
+            although this function is never used since 0.7.0 but
+            libmenu-cache always requests exact socket name instead */
     g_snprintf( buf, len, "%s/.menu-cached-%s-%s", g_get_tmp_dir(),
                 dpy ? dpy : ":0", g_get_user_name() );
     g_free(dpy);
 }
 
-static int create_socket(struct sockaddr_un *addr)
+static gboolean socket_is_alive(struct sockaddr_un *addr)
 {
-    int fd = -1;
+    int fd = socket(PF_UNIX, SOCK_STREAM, 0);
+    socklen_t len = sizeof(sa_family_t) + strlen(addr->sun_path) + 1;
 
-    fd = socket(PF_UNIX, SOCK_STREAM, 0);
     if (fd < 0)
     {
         DEBUG("Failed to create socket");
-        return -1;
+        /* still return TRUE to drop attempt */
+        return TRUE;
     }
+    if (connect(fd, (struct sockaddr*)addr, len) >= 0)
+    {
+        /* there is a listener on the socket */
+        close(fd);
+        DEBUG("Another menu-cached seems to reside on the socket");
+        return TRUE;
+    }
+    close(fd);
 
     /* remove previous socket file */
     if (unlink(addr->sun_path) < 0) {
@@ -497,19 +538,57 @@ static int create_socket(struct sockaddr_un *addr)
     /* remove of previous socket file successful */
     else
         g_warning("removed previous socket file %s", addr->sun_path);
+    return FALSE;
+}
 
-    if(bind(fd, (struct sockaddr *)addr, sizeof(*addr)) < 0)
+static int create_socket(struct sockaddr_un *addr)
+{
+    int fd = -1;
+    char *lockfile;
+
+    lockfile = g_strconcat(addr->sun_path, ".lock", NULL);
+    fd = open(lockfile, O_CREAT | O_EXCL | O_WRONLY, 0700);
+    if (fd < 0)
+    {
+        DEBUG("Cannot create lock file %s: %s", lockfile, strerror(errno));
+        g_free(lockfile);
+        return -1;
+    }
+    close(fd);
+
+    fd = socket(PF_UNIX, SOCK_STREAM, 0);
+    if (fd < 0)
+    {
+        DEBUG("Failed to create socket");
+    }
+
+    /* remove previous socket file */
+    else if (g_file_test(addr->sun_path, G_FILE_TEST_EXISTS) &&
+             socket_is_alive(addr))
+    {
+        close(fd);
+        fd = -1;
+    }
+
+    else if(bind(fd, (struct sockaddr *)addr, sizeof(*addr)) < 0)
     {
         DEBUG("Failed to bind to socket");
         close(fd);
-        return -1;
+        fd = -1;
     }
-    if(listen(fd, 30) < 0)
+
+    else if(listen(fd, 30) < 0)
     {
         DEBUG( "Failed to listen to socket" );
         close(fd);
-        return -1;
+        fd = -1;
     }
+
+    else
+        fcntl(fd, F_SETFD, FD_CLOEXEC);
+
+    unlink(lockfile);
+    g_free(lockfile);
     return fd;
 }
 
@@ -717,6 +796,14 @@ retry:
     return ret;
 }
 
+static void terminate(int sig)
+{
+/* #ifndef HAVE_ABSTRACT_SOCKETS */
+    unlink(socket_file);
+    exit(0);
+/* #endif */
+}
+
 static gboolean on_new_conn_incoming(GIOChannel* ch, GIOCondition cond, gpointer user_data)
 {
     int server, client;
@@ -732,9 +819,16 @@ static gboolean on_new_conn_incoming(GIOChannel* ch, GIOCondition cond, gpointer
     if( client == -1 )
     {
         DEBUG("accept error");
-        return TRUE;
+        if (errno == ECONNABORTED)
+            /* client failed, just continue */
+            return TRUE;
+        /* else it's socket error, terminate server */
+        terminate(SIGTERM);
+        return FALSE;
     }
 
+    fcntl (client, F_SETFD, FD_CLOEXEC);
+
     child = g_io_channel_unix_new(client);
     g_io_channel_set_close_on_unref( child, TRUE );
 
@@ -749,14 +843,6 @@ static gboolean on_new_conn_incoming(GIOChannel* ch, GIOCondition cond, gpointer
     return TRUE;
 }
 
-static void terminate(int sig)
-{
-/* #ifndef HAVE_ABSTRACT_SOCKETS */
-    unlink(socket_file);
-    exit(0);
-/* #endif */
-}
-
 static gboolean on_server_conn_close(GIOChannel* ch, GIOCondition cond, gpointer user_data)
 {
     /* FIXME: is this possible? */
@@ -771,12 +857,48 @@ int main(int argc, char** argv)
     int server_fd;
     struct sockaddr_un addr;
 #ifndef DISABLE_DAEMONIZE
-    int fd, pid;
+    pid_t pid;
+    int status;
 
     long open_max;
     long i;
 #endif
 
+#ifndef DISABLE_DAEMONIZE
+    /* Become a daemon */
+    if ((pid = fork()) < 0) {
+        g_error("can't fork");
+        return 2;
+    }
+    else if (pid != 0) {
+        /* wait for result of child */
+        while ((i = waitpid(pid, &status, 0)) < 0 && errno == EINTR);
+        if (i < 0 || !WIFEXITED(status)) /* system error or child crashed */
+            return 127;
+        /* exit parent */
+        return WEXITSTATUS(status);
+    }
+
+    /* reset session to forget about parent process completely */
+    setsid();
+
+    /* change working directory to root, so previous working directory
+     * can be unmounted */
+    if (chdir("/") < 0) {
+        g_error("can't change directory to /");
+    }
+
+    /* don't hold open fd opened besides server socket and std{in,out,err} */
+    open_max = sysconf (_SC_OPEN_MAX);
+    for (i = 3; i < open_max; i++)
+        close (i);
+
+    /* /dev/null for stdin, stdout, stderr */
+    if (freopen("/dev/null", "r", stdin)) i = i;
+    if (freopen("/dev/null", "w", stdout)) i = i;
+    if (freopen("/dev/null", "w", stderr)) i = i;
+#endif
+
     memset(&addr, 0, sizeof(addr));
     addr.sun_family = AF_UNIX;
     if (argc < 2)
@@ -797,46 +919,17 @@ int main(int argc, char** argv)
         return 1;
 
 #ifndef DISABLE_DAEMONIZE
-    /* Become a daemon */
+    /* Second fork to let parent get a result */
     if ((pid = fork()) < 0) {
         g_error("can't fork");
+        close(server_fd);
+        unlink(socket_file);
+        return 2;
+    } else if (pid != 0) {
+        /* exit child */
+        return 0;
     }
-    else if (pid != 0) {
-        /* exit parent */
-        exit(0);
-    }
-
-    /* reset session to forget about parent process completely */
-    setsid();
-
-    /* change working directory to root, so previous working directory
-     * can be unmounted */
-    if (chdir("/") < 0) {
-        g_error("can't change directory to /");
-    }
-
-    open_max = sysconf (_SC_OPEN_MAX);
-    for (i = 0; i < open_max; i++)
-    {
-        /* don't hold open fd opened besides server socket */
-        if (i != server_fd)
-            fcntl (i, F_SETFD, FD_CLOEXEC);
-    }
-
-    /* /dev/null for stdin, stdout, stderr */
-    fd = open ("/dev/null", O_RDONLY);
-    if (fd != -1)
-    {
-        dup2 (fd, 0);
-        close (fd);
-    }
-    fd = open ("/dev/null", O_WRONLY);
-    if (fd != -1)
-    {
-        dup2 (fd, 1);
-        dup2 (fd, 2);
-        close (fd);
-    }
+    /* We are in grandchild now, daemonized */
 #endif
 
     signal(SIGHUP, terminate);
